
__version__ = '0.1.4'

import argparse
import logging
import pathlib
import sys

from .asset import font, image, map, raw
from .tool import cmake, flasher, metadata, packer, relocs


def exception_handler(exception_type, exception, traceback):
    sys.stderr.write(f"{type(exception).__name__}: {exception}\n")
    sys.stderr.flush()


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--debug', action='store_true', help='Enable exception traces')
    parser.add_argument('-v', '--verbosity', action='count', default=0, help='Output verbosity')
    subparsers = parser.add_subparsers(dest='command', help='Commands')

    if len(sys.argv) == 2:
        file = pathlib.Path(sys.argv[1])
        if file.is_file() and file.exists():
            f = flasher.Flasher(subparsers)
            f.run_save({'file': file, 'directory': '/'})
            return

    tools = {}

    tools[image.ImageAsset.command] = image.ImageAsset(subparsers)
    tools[font.FontAsset.command] = font.FontAsset(subparsers)
    tools[map.MapAsset.command] = map.MapAsset(subparsers)
    tools[raw.RawAsset.command] = raw.RawAsset(subparsers)

    # Add the non-asset tools
    tools[packer.Packer.command] = packer.Packer(subparsers)
    tools[cmake.CMake.command] = cmake.CMake(subparsers)
    tools[flasher.Flasher.command] = flasher.Flasher(subparsers)
    tools[metadata.Metadata.command] = metadata.Metadata(subparsers)
    tools[relocs.Relocs.command] = relocs.Relocs(subparsers)

    args = parser.parse_args()

    log_format = '%(levelname)s: %(message)s'

    log_verbosity = min(args.verbosity, 3)
    log_level = [logging.ERROR, logging.WARNING, logging.INFO, logging.DEBUG][log_verbosity]
    logging.basicConfig(level=log_level, format=log_format)

    if args.debug:
        logging.basicConfig(level=logging.DEBUG, format=log_format)
    else:
        sys.excepthook = exception_handler

    if args.command is None:
        parser.print_help()
    else:
        sys.exit(tools[args.command].run(args))



from ..core.assetbuilder import AssetBuilder


class RawAsset(AssetBuilder):
    command = 'raw'
    help = 'Convert raw/binary or csv data for 32Blit'
    types = ['binary', 'csv']
    typemap = {
        'binary': ('.bin', '.raw'),
        'csv': ('.csv')
    }

    def prepare(self, args):
        AssetBuilder.prepare(self, args)

    def csv_to_binary(self, input_data, base=10, offset=0):
        try:
            input_data = input_data.decode('utf-8')
        except AttributeError:
            pass

        input_data = input_data.strip()

        # Replace '1, 2, 3' to '1,2,3', might as well do it here
        input_data = input_data.replace(' ', '')

        # Split out into rows on linebreak
        input_data = input_data.split('\n')

        # Split every row into columns on the comma
        input_data = [row.split(',') for row in input_data]

        # Flatten our rows/cols 2d array into a 1d array of bytes
        # Might as well do the int conversion here, to save another loop
        input_data = [(int(col, base) + offset) for row in input_data for col in row if col != '']

        return bytes(input_data)

    def to_binary(self, input_data):
        if self.input_type == 'csv':
            input_data = self.csv_to_binary(input_data, base=10)
        return input_data



import binascii
import math

from construct import (Adapter, Array, Bytes, Checksum, Computed, Const,
                       Int8ul, Int16ul, Int32ub, Int32ul, Optional,
                       PaddedString, Prefixed, PrefixedArray, RawCopy, Rebuild,
                       Struct, len_, this)


def compute_bit_length(ctx):
    """Compute the required bit length for image data.
    Uses the count of items in the palette to determine how
    densely we can pack the image data.
    """
    if ctx.type == "RW":
        return 8
    else:
        return max(1, (ctx.palette_entries - 1).bit_length())


def compute_data_length(ctx):
    """Compute the required data length for palette based images.
    We need this computation here so we can use `math.ceil` and
    byte-align the result.
    """
    return math.ceil((ctx.width * ctx.height * ctx.bit_length) / 8)


class PaletteCountAdapter(Adapter):
    def _decode(self, obj, context, path):
        if obj == 0:
            obj = 256
        return obj

    def _encode(self, obj, context, path):
        if obj == 256:
            obj = 0
        return obj


struct_blit_pixel = Struct(
    'r' / Int8ul,
    'g' / Int8ul,
    'b' / Int8ul,
    'a' / Int8ul
)

struct_blit_image = Struct(
    'header' / Const(b'SPRITE'),
    'type' / PaddedString(2, 'ASCII'),
    'size' / Rebuild(Int32ul, len_(this.data) + (this.palette_entries * 4) + 18),
    'width' / Int16ul,
    'height' / Int16ul,
    'format' / Const(0x02, Int8ul),
    'palette_entries' / PaletteCountAdapter(Int8ul),
    'palette' / Array(this.palette_entries, struct_blit_pixel),
    'bit_length' / Computed(compute_bit_length),
    'data_length' / Computed(compute_data_length),
    'data' / Array(this.data_length, Int8ul)
)

struct_blit_meta = Struct(
    'header' / Const(b'BLITMETA'),
    'data' / Prefixed(Int16ul, Struct(
        'checksum' / Checksum(
            Int32ul,
            lambda data: binascii.crc32(data),
            this._._.bin.data
        ),
        'date' / PaddedString(16, 'ascii'),
        'title' / PaddedString(25, 'ascii'),
        'description' / PaddedString(129, 'ascii'),
        'version' / PaddedString(17, 'ascii'),
        'author' / PaddedString(17, 'ascii'),
        'icon' / struct_blit_image,
        'splash' / struct_blit_image
    ))
)

struct_blit_meta_standalone = Struct(
    'header' / Const(b'BLITMETA'),
    'data' / Prefixed(Int16ul, Struct(
        'checksum' / Int32ul,
        'date' / PaddedString(16, 'ascii'),
        'title' / PaddedString(25, 'ascii'),
        'description' / PaddedString(129, 'ascii'),
        'version' / PaddedString(17, 'ascii'),
        'author' / PaddedString(17, 'ascii'),
        'icon' / struct_blit_image,
        'splash' / struct_blit_image
    ))
)

struct_blit_bin = Struct(
    'header' / Const(b'BLIT'),
    'render' / Int32ul,
    'update' / Int32ul,
    'init' / Int32ul,
    'length' / Int32ul,
    # The length above is actually the _flash_end symbol from startup_user.s
    # it includes the offset into 0x90000000 (external flash)
    # we mask out the highest nibble to correct this into the actual bin length
    # plus subtract 20 bytes for header, symbol and length dwords
    'bin' / Bytes((this.length & 0x0FFFFFFF) - 20)
)

struct_blit_relo = Struct(
    'header' / Const(b'RELO'),
    'relocs' / PrefixedArray(Int32ul, Struct(
        'reloc' / Int32ul
    ))
)

blit_game = Struct(
    'relo' / Optional(struct_blit_relo),
    'bin' / RawCopy(struct_blit_bin),
    'meta' / Optional(struct_blit_meta)
)

blit_game_with_meta = Struct(
    'relo' / Optional(struct_blit_relo),
    'bin' / RawCopy(struct_blit_bin),
    'meta' / struct_blit_meta
)

blit_game_with_meta_and_relo = Struct(
    'relo' / struct_blit_relo,
    'bin' / RawCopy(struct_blit_bin),
    'meta' / struct_blit_meta
)


def compute_icns_data_length(ctx):
    """Compute the required data length for palette based images.
    We need this computation here so we can use `math.ceil` and
    byte-align the result.
    """
    return math.ceil((ctx.width * ctx.height * ctx.bit_length) / 8)


blit_icns = Struct(
    'header' / Const(b'icns'),
    'size' / Rebuild(Int32ub, len_(this.data) + 16),
    'type' / Const(b'ic07'),  # 128×128 icon in PNG format
    'data_length' / Rebuild(Int32ub, len_(this.data) + 8),
    'data' / Bytes(this.data_length - 8)
)



import logging
import pathlib

import yaml

from ..asset import font, image, map, raw
from ..core.outputformat import CHeader, CSource, RawBinary
from ..core.palette import Palette
from ..core.tool import Tool


class Packer(Tool):
    command = 'pack'
    help = 'Pack a collection of assets for 32Blit'

    def __init__(self, parser):
        Tool.__init__(self, parser)
        self.parser.add_argument('--config', type=pathlib.Path, help='Asset config file')
        self.parser.add_argument('--output', type=pathlib.Path, help='Name for output file(s) or root path when using --config')
        self.parser.add_argument('--files', nargs='+', type=pathlib.Path, help='Input files')
        self.parser.add_argument('--force', action='store_true', help='Force file overwrite')

        self.config = {}
        self.assets = []
        self.general_options = {}

    def parse_config(self, config_file):
        config = open(config_file).read()
        config = yaml.safe_load(config)

        self.config = config

    def filelist_to_config(self, filelist, output):
        config = {}

        for file in filelist:
            config[file] = {}

        self.config = {f'{output}': config}

    def get_general_options(self):
        for key, value in self.config.items():
            if key in ():
                self.general_options[key] = value

        for key, value in self.general_options.items():
            self.config.items.pop(key)

    def run(self, args):
        self.working_path = pathlib.Path('.')

        if args.config is not None:
            if args.config.is_file():
                self.working_path = args.config.parent
            else:
                logging.warning(f'Unable to find config at {args.config}')

        if args.output is not None:
            self.destination_path = args.output
        else:
            self.destination_path = self.working_path

        if args.config is not None:
            self.parse_config(args.config)
            logging.info(f'Using config at {args.config}')

        elif args.files is not None and args.output is not None:
            self.filelist_to_config(args.files, args.output)

        self.get_general_options()

        # Top level of our config is filegroups and general settings
        for target, options in self.config.items():
            output_file = self.working_path / target
            logging.info(f'Preparing output target {output_file}')

            asset_sources = []
            target_options = {}

            for key, value in options.items():
                if key in AssetTarget.supported_options:
                    target_options[key] = value

            # Strip high-level options from the dict
            # Leaving just file source globs
            for key, value in target_options.items():
                options.pop(key)

            for file_glob, file_options in options.items():
                # Treat the input string as a glob, and get an input filelist
                if type(file_glob) is str:
                    input_files = list(self.working_path.glob(file_glob))
                else:
                    input_files = [file_glob]
                if len(input_files) == 0:
                    logging.warning(f'Input file(s) not found {self.working_path / file_glob}')
                    continue

                # Rewrite a single string option to `name: option`
                # This handles: `inputfile.bin: filename` entries
                if type(file_options) is str:
                    file_options = {'name': file_options}

                elif file_options is None:
                    file_options = {}

                asset_sources.append(
                    AssetSource(input_files, file_options=file_options, working_path=self.working_path)
                )

            self.assets.append(AssetTarget(
                output_file,
                asset_sources,
                target_options
            ))

        self.destination_path.mkdir(parents=True, exist_ok=True)

        for target in self.assets:
            output = target.build()
            self.output(output, self.destination_path / target.output_file.name, target.output_format, force=args.force)


class AssetSource():
    supported_options = ('name', 'type')

    def __init__(self, input_files, file_options, working_path):
        self.asset_builders = {
            image.ImageAsset.command: image.ImageAsset,
            font.FontAsset.command: font.FontAsset,
            map.MapAsset.command: map.MapAsset,
            raw.RawAsset.command: raw.RawAsset
        }

        self.input_files = input_files
        self.working_path = working_path
        self.builder_options = {}
        self.type = None
        self.name = None
        self.parse_source_options(file_options)

    def parse_source_options(self, opts):
        for option, value in opts.items():
            if option not in self.supported_options:
                # These options are passed into the Map/Image/Raw builder
                self.builder_options[option] = value
            else:
                setattr(self, option, value)

        if self.type is None:
            # Glob files all have the same suffix, so we only care about the first one
            self.type = self.guess_type(self.input_files[0])

    def guess_type(self, file):
        # We need a high-level guess type that can query the types each builder supports
        # and dispatch our builds to the right one
        for command, asset_builder in self.asset_builders.items():
            for input_type, suffixes in asset_builder.typemap.items():
                if file.suffix in suffixes:
                    return f'{command}/{input_type}'
        logging.warning(f'Unable to guess type, assuming raw/binary {file}')
        return 'raw/binary'

    def build(self, format, prefix=None, output_file=None):
        builder, input_type = self.type.split('/')
        builder = self.asset_builders[builder]()

        # Now we know our target builder, one last iteration through the options
        # allows some pre-processing stages to remap paths or other idiosyncrasies
        # of the yml config format.
        for option_name, option_value in builder.options.items():
            if option_name in self.builder_options:
                option_type = builder.options[option_name]
                if type(option_type) is tuple:
                    option_type, default_value = option_type

                # Ensure Palette and pathlib.Path type options for this asset builder
                # are prefixed with the working directory if they are not absolute paths
                if option_type in (Palette, pathlib.Path):
                    if not pathlib.Path(self.builder_options[option_name]).is_absolute():
                        self.builder_options[option_name] = self.working_path / self.builder_options[option_name]

        output = []

        for file in self.input_files:
            self.builder_options.update({
                'input_file': file,
                'output_file': output_file,
                'input_type': input_type,
                'output_format': format,
                'symbol_name': self.name,
                'working_path': self.working_path,
                'prefix': prefix
            })
            builder.prepare_options(self.builder_options)
            output.append(builder.build())
            logging.info(f' - {self.type} {file} -> {builder.symbol_name}')

        return output


class AssetTarget():
    supported_options = ('prefix', 'type')
    output_formats = {
        '.hpp': CHeader,
        '.cpp': CSource,
        '.blit': RawBinary,
        '.raw': RawBinary
    }

    def __init__(self, output_file, sources, target_options):
        self.output_file = output_file
        self.sources = sources
        self.parse_target_options(target_options)

    def parse_target_options(self, target_options):
        self.prefix = target_options.get('prefix')
        output_format = target_options.get('type')
        if output_format is None:
            output_format = self.guess_output_format(self.output_file)
        else:
            if type not in self.output_formats.values():
                raise ValueError(f'Unknown asset output format {output_format}')
        self.output_format = output_format

    def build(self):
        # TODO I hate this code, but because we're dealing with multiple input files
        # the output can potentially be spread across multiple files with multiple snippets
        # We need to gather each input files declarations/definitions (or otherwise) into
        # the right output file.
        output = []

        # If the output_format deals with multiple files, switch to a dict and prep the files
        if self.output_format.components is not None:
            output = {}
            for ext in self.output_format.components:
                output[ext] = []

        # Iterate through all the sources (input files) and build them
        for source in self.sources:
            data = source.build(format=self.output_format, output_file=self.output_file, prefix=self.prefix)
            for snippet in data:
                if self.output_format.components is not None:
                    # If there are multiple output files, collect all the snippets into their respective outputs
                    for ext, content in snippet.items():
                        output[ext].append(content)
                else:
                    # If it's a single file format just append the snippet to the list
                    output.append(snippet)

        return output

    def guess_output_format(self, file):
        if file.suffix in self.output_formats:
            return self.output_formats[file.suffix]
        logging.warning(f'Unable to guess type of {file}, assuming raw/binary')
        return RawBinary



from .raw import RawAsset


class MapAsset(RawAsset):
    command = 'map'
    help = 'Convert popular tilemap formats for 32Blit'
    types = ['tiled']
    typemap = {
        'tiled': ('.tmx', '.raw'),
    }

    def tiled_to_binary(self, input_data):
        from xml.etree import ElementTree as ET
        root = ET.fromstring(input_data)
        layers = root.findall('layer/data')
        data = []
        for layer in layers:
            try:
                data.append(self.csv_to_binary(
                    layer.text,
                    base=10,
                    offset=-1))  # Tiled indexes from 1 rather than 0
            except ValueError:
                raise RuntimeError("Failed to convert .tmx, does it contain blank (0) tiles? Tiled is 1-indexed, so these get converted to -1 and blow everyting up")
        return b''.join(data)

    def to_binary(self, input_data):
        if self.input_type == 'tiled':
            return self.tiled_to_binary(input_data)



import logging


class Tool():
    options = {}

    def __init__(self, parser=None):
        self.parser = None
        if parser is not None:
            self.parser = parser.add_parser(self.command, help=self.help)

    def prepare(self, opts):
        for option, option_type in self.options.items():
            default_value = None
            if type(option_type) is tuple:
                option_type, default_value = option_type
            setattr(self, option, opts.get(option, default_value))

    def run(self, args):
        raise NotImplementedError

    def output(self, output_data, output_file, output_format, force=False):
        if output_file is None:
            output_data = output_format().join(None, output_file, output_data)
            print(output_data)
        else:
            if type(output_data) is dict:
                for extension, data in output_data.items():
                    data = output_format().join(extension, output_file, data)
                    self.write_file(output_file.with_suffix(f'.{extension}'), data, force)
            else:
                output_data = output_format().join(None, output_file, output_data)
                self.write_file(output_file, output_data, force)

    def write_file(self, output_file, output_data, force=False):
        if output_file.exists() and not force:
            raise ValueError(f'Refusing to overwrite {output_file} (use force)')
        else:
            logging.info(f'Writing {output_file}')
            if type(output_data) is str:
                open(output_file, 'wb').write(output_data.encode('utf-8'))
            else:
                open(output_file, 'wb').write(output_data)



import io
import struct

import freetype
from PIL import Image

from ..core.assetbuilder import AssetBuilder


class FontAsset(AssetBuilder):
    command = 'font'
    help = 'Convert fonts for 32Blit'
    types = ['image', 'font']
    typemap = {
        'image': ('.png', '.gif'),
        'font': ('.ttf')  # possibly other freetype supported formats...
    }

    def __init__(self, parser=None):
        self.options.update({
            'height': (int, 0),
            'horizontal_spacing': (int, 1),
            'vertical_spacing': (int, 1),
            'space_width': (int, 3)
        })

        AssetBuilder.__init__(self, parser)

        self.height = 0
        self.horizontal_spacing = 1
        self.vertical_spacing = 1
        self.space_width = 3
        self.base_char = ord(' ')
        self.num_chars = 96

        if self.parser is not None:
            self.parser.add_argument('--height', type=int, default=0, help='Font height (calculated from image if not specified)')
            self.parser.add_argument('--horizontal-spacing', type=int, default=1, help='Additional space between characters for variable-width mode')
            self.parser.add_argument('--vertical-spacing', type=int, default=1, help='Space between lines')
            self.parser.add_argument('--space-width', type=int, default=1, help='Width of the space character')

    def prepare(self, args):
        AssetBuilder.prepare(self, args)

    def process_image_font(self, input_data):
        # Since we already have bytes, we need to pass PIL an io.BytesIO object
        image = Image.open(io.BytesIO(input_data)).convert('1')
        w, h = image.size

        if self.height != 0 and self.height != h:
            raise TypeError("Specified height does not match image height")

        char_width = w // self.num_chars
        char_height = h

        font_data = []
        font_w = []  # per character width for variable-width mode

        for c in range(0, self.num_chars):
            char_w = 0

            for x in range(0, char_width):
                byte = 0

                for y in range(0, h):
                    bit = y % 8

                    # next byte
                    if bit == 0 and y > 0:
                        font_data.append(byte)
                        byte = 0

                    if image.getpixel((x + c * char_width, y)) != 0:
                        byte |= 1 << bit
                        if x + 1 > char_w:
                            char_w = x + 1

                font_data.append(byte)

            if c == 0:  # space
                font_w.append(self.space_width)
            else:
                font_w.append(char_w + self.horizontal_spacing)

        return font_data, font_w, char_width, char_height

    def process_ft_font(self, input_data):
        if self.height == 0:
            raise TypeError("Height must be specified for font files")

        face = freetype.Face(io.BytesIO(input_data))

        # request height
        face.set_pixel_sizes(0, self.height)

        char_width = 0
        char_height = 0
        min_y = self.height
        font_w = []

        # measure the actual size of the characters (may not match requested)
        for c in range(0, self.num_chars):
            face.load_char(c + self.base_char, freetype.FT_LOAD_RENDER | freetype.FT_LOAD_TARGET_MONO)

            font_w.append(face.glyph.advance.x >> 6)

            if face.glyph.bitmap.width + face.glyph.bitmap_left > char_width:
                char_width = face.glyph.bitmap.width + face.glyph.bitmap_left

            if (self.height - face.glyph.bitmap_top) + face.glyph.bitmap.rows > char_height:
                char_height = self.height - face.glyph.bitmap_top + face.glyph.bitmap.rows

            if self.height - face.glyph.bitmap_top < min_y:
                min_y = self.height - face.glyph.bitmap_top

        char_height -= min_y  # trim empty space at the top

        font_data = []

        # now do the conversion
        for c in range(0, self.num_chars):
            face.load_char(c + self.base_char, freetype.FT_LOAD_RENDER | freetype.FT_LOAD_TARGET_MONO)

            x_off = face.glyph.bitmap_left
            y_off = self.height - face.glyph.bitmap_top - min_y

            for x in range(0, char_width):
                byte = 0

                for y in range(0, char_height):
                    bit = y % 8

                    # next byte
                    if bit == 0 and y > 0:
                        font_data.append(byte)
                        byte = 0

                    if x < x_off or x - x_off >= face.glyph.bitmap.width or y < y_off or y - y_off >= face.glyph.bitmap.rows:
                        continue

                    # freetype monochrome bitmaps are the other way around
                    if face.glyph.bitmap.buffer[(x - x_off) // 8 + (y - y_off) * face.glyph.bitmap.pitch] & (0x80 >> ((x - x_off) & 7)):
                        byte |= 1 << bit

                font_data.append(byte)

        return font_data, font_w, char_width, char_height

    def to_binary(self, input_data):
        if self.input_type == 'image':
            font_data, font_w_data, char_width, char_height = self.process_image_font(input_data)
        elif self.input_type == 'font':
            font_data, font_w_data, char_width, char_height = self.process_ft_font(input_data)

        font_data = bytes(font_data)

        head_data = struct.pack('<BBBB', self.num_chars, char_width, char_height, self.vertical_spacing)

        data = bytes('FONT', encoding='utf-8')
        data += head_data
        data += bytes(font_w_data)
        data += bytes(font_data)

        return data



import logging
import pathlib

import yaml

from ..core.tool import Tool
from ..tool.packer import AssetTarget


class CMake(Tool):
    command = 'cmake'
    help = 'Generate CMake configuration for the asset packer'

    def __init__(self, parser):
        Tool.__init__(self, parser)
        self.parser.add_argument('--config', type=pathlib.Path, help='Asset config file')
        self.parser.add_argument('--cmake', type=pathlib.Path, help='Output CMake file', required=True)
        self.parser.add_argument('--output', type=pathlib.Path, help='Name for output file(s) or root path when using --config')

        self.config = {}
        self.general_options = {}

    def parse_config(self, config_file):
        config = open(config_file).read()
        config = yaml.safe_load(config)

        self.config = config

    def get_general_options(self):
        for key, value in self.config.items():
            if key in ():
                self.general_options[key] = value

        for key, value in self.general_options.items():
            self.config.items.pop(key)

    def run(self, args):
        self.working_path = pathlib.Path('.')

        if args.config is not None:
            if args.config.is_file():
                self.working_path = args.config.parent
            else:
                logging.warning(f'Unable to find config at {args.config}')

        if args.config is not None:
            self.parse_config(args.config)
            logging.info(f'Using config at {args.config}')

        if args.output is not None:
            self.destination_path = args.output
        else:
            self.destination_path = self.working_path

        if 'title' in self.config and 'description' in self.config:
            logging.info('Detected metadata config')
            self.run_for_metadata_config(args)
        else:
            logging.info('Detected asset config')
            self.run_for_asset_config(args)

    def run_for_metadata_config(self, args):
        all_inputs = []

        if 'splash' in self.config:
            file = pathlib.Path(self.config['splash']['file'])
            if file.is_absolute():
                all_inputs += [file]
            else:
                all_inputs += list(self.working_path.glob(str(file)))

        if 'icon' in self.config:
            file = pathlib.Path(self.config['icon']['file'])
            if file.is_absolute():
                all_inputs += [file]
            else:
                all_inputs += list(self.working_path.glob(str(file)))

        all_inputs = '\n'.join(f'"{x}"'.replace('\\', '/') for x in all_inputs)

        open(args.cmake, 'w').write(f'''# Auto Generated File - DO NOT EDIT!
    set(METADATA_DEPENDS
    {all_inputs}
    )
    set(METADATA_TITLE "{self.config['title']}")
    set(METADATA_AUTHOR "{self.config['author']}")
    set(METADATA_DESCRIPTION "{self.config['description']}")
    set(METADATA_VERSION "{self.config['version']}")
    ''')

    def run_for_asset_config(self, args):
        self.get_general_options()

        all_inputs = []
        all_outputs = []

        # Top level of our config is filegroups and general settings
        for target, options in self.config.items():
            target = pathlib.Path(target)

            if target.suffix in AssetTarget.output_formats:
                output_formatter = AssetTarget.output_formats[target.suffix]
            else:
                logging.warning(f'Unable to guess type of {target}, assuming raw/binary')
                output_formatter = AssetTarget.output_formats['.raw']

            if output_formatter.components is None:
                all_outputs.append(self.destination_path / target.name)
            else:
                for suffix in output_formatter.components:
                    all_outputs.append(self.destination_path / target.with_suffix(f'.{suffix}').name)

            # Strip high-level options from the dict
            # Leaving just file source globs
            for key in AssetTarget.supported_options:
                options.pop(key, None)

            for file_glob, file_options in options.items():
                # Rewrite a single string option to `name: option`
                # This handles: `inputfile.bin: filename` entries
                if type(file_options) is str:
                    file_options = {'name': file_options}

                input_files = []
                # Parse the options for any references to input files
                for key, value in file_options.items():
                    if key in ('palette') and type(value) is str:
                        input_files += list(self.working_path.glob(value))
                # Treat the input string as a glob, and get an input filelist
                if type(file_glob) is str:
                    input_files += list(self.working_path.glob(file_glob))
                else:
                    input_files += [file_glob]
                if len(input_files) == 0:
                    logging.warning(f'Input file(s) not found {self.working_path / file_glob}')
                    continue

                all_inputs += input_files

        all_inputs = '\n'.join(f'"{x}"'.replace('\\', '/') for x in all_inputs)
        all_outputs = '\n'.join(f'"{x}"'.replace('\\', '/') for x in all_outputs)

        open(args.cmake, 'w').write(f'''# Auto Generated File - DO NOT EDIT!
set(ASSET_DEPENDS
{all_inputs}
)

set(ASSET_OUTPUTS
{all_outputs}
)
''')



import logging
import pathlib
import struct
import time

import serial.tools.list_ports
from construct.core import ConstructError
from serial.serialutil import SerialException
from tqdm import tqdm

from ..core.struct import struct_blit_meta_standalone
from ..core.tool import Tool


class Flasher(Tool):
    command = 'flash'
    help = 'Flash a binary or save games/files to 32Blit'

    def __init__(self, subparser):
        Tool.__init__(self, subparser)

        self.parser.add_argument('--port', help='Serial port', type=self.validate_comport)

        operations = self.parser.add_subparsers(dest='operation', help='Flasher operations')

        self.op_save = operations.add_parser('save', help='Save a game/file to your 32Blit')
        self.op_save.add_argument('--file', type=pathlib.Path, required=True, help='File to save')
        self.op_save.add_argument('--directory', type=str, default='/', help='Target directory')

        self.op_flash = operations.add_parser('flash', help='Flash a game to your 32Blit')
        self.op_flash.add_argument('--file', type=pathlib.Path, required=True, help='File to flash')

        self.op_delete = operations.add_parser('delete', help='Delete a game/file from your 32Blit')
        group = self.op_delete.add_mutually_exclusive_group(required=True)
        group.add_argument('--offset', type=int, help='Flash offset of game to delete')
        group.add_argument('--block', type=int, help='Flash block of game to delete')

        self.op_list = operations.add_parser('list', help='List games/files on your 32Blit')
        self.op_debug = operations.add_parser('debug', help='Enter serial debug mode')
        self.op_reset = operations.add_parser('reset', help='Reset your 32Blit')
        self.op_info = operations.add_parser('info', help='Get 32Blit run status')

    def find_comport(self):
        ret = []
        for comport in serial.tools.list_ports.comports():
            if comport.vid == 0x0483 and comport.pid == 0x5740:
                logging.info(f'Found 32Blit on {comport.device}')
                ret.append(comport.device)

        if ret:
            return ret

        raise RuntimeError('Unable to find 32Blit')

    def validate_comport(self, device):
        if device.lower() == 'auto':
            return self.find_comport()[:1]
        if device.lower() == 'all':
            return self.find_comport()

        for comport in serial.tools.list_ports.comports():
            if comport.device == device:
                if comport.vid == 0x0483 and comport.pid == 0x5740:
                    logging.info(f'Found 32Blit on {comport.device}')
                    return [device]
        raise RuntimeError(f'Unable to find 32Blit on {device}')

    def run(self, args):
        if args.operation is not None:
            dispatch = f'run_{args.operation}'
            getattr(self, dispatch)(vars(args))

    def serial_command(fn):
        """Set up and tear down serial connections."""
        def _decorated(self, args):
            ports = args.get('port', None)
            if ports is None:
                ports = self.find_comport()

            for port in ports:
                sp = serial.Serial(port, timeout=5)
                fn(self, sp, args)
                sp.close()
        return _decorated

    def _get_status(self, serial):
        serial.write(b'32BLINFO\x00')
        response = serial.read(8)
        if response == b'':
            raise RuntimeError('Timeout waiting for 32Blit status.')
        return 'game' if response == b'32BL_EXT' else 'firmware'

    def _reset(self, serial, timeout=5.0):
        serial.write(b'32BL_RST\x00')
        serial.close()
        t_start = time.time()
        while time.time() - t_start < timeout:
            try:
                serial.open()
                break
            except SerialException:
                time.sleep(0.1)

    def _reset_to_firmware(self, serial):
        if self._get_status(serial) == 'game':
            self._reset(serial)

    def _send_file(self, serial, file, dest, directory='/'):
        sent_byte_count = 0
        chunk_size = 64
        file_name = file.name
        file_size = file.stat().st_size

        if dest == 'sd':
            if not directory.endswith('/'):
                directory = f'{directory}/'

            logging.info(f'Saving {file} ({file_size} bytes) as {file_name} in {directory}')
            command = f'32BLSAVE{directory}{file_name}\x00{file_size}\x00'
        elif dest == 'flash':
            logging.info(f'Flashing {file} ({file_size} bytes)')
            command = f'32BLPROG{file_name}\x00{file_size}\x00'

        serial.reset_output_buffer()
        serial.write(command.encode('ascii'))

        with open(file, 'rb') as file:
            progress = tqdm(total=file_size, desc="Flashing...", unit_scale=True, unit_divisor=1024, unit="B", ncols=70, dynamic_ncols=True)

            while sent_byte_count < file_size:
                data = file.read(chunk_size)
                serial.write(data)
                sent_byte_count += chunk_size
                progress.update(chunk_size)

    @serial_command
    def run_save(self, serial, args):
        self._reset_to_firmware(serial)
        self._send_file(serial, args.get('file'), 'sd', directory=args.get('directory'))

    @serial_command
    def run_flash(self, serial, args):
        self._reset_to_firmware(serial)
        self._send_file(serial, args.get('file'), 'flash')

    @serial_command
    def run_delete(self, serial, args):
        self._reset_to_firmware(serial)
        serial.write(b'32BLERSE\x00')

        offset = args.get('offset')
        if offset is None:
            offset = args.get('block') * 64 * 1024

        serial.write(struct.pack("<I", offset))

    @serial_command
    def run_list(self, serial, args):
        self._reset_to_firmware(serial)

        serial.write(b'32BL__LS\x00')
        offset_str = serial.read(4)

        while offset_str != '' and offset_str != b'\xff\xff\xff\xff':
            offset, = struct.unpack('<I', offset_str)

            size, = struct.unpack('<I', serial.read(4))
            meta_head = serial.read(10)
            meta_size, = struct.unpack('<H', meta_head[8:])

            meta = None
            if meta_size:
                size += meta_size + 10
                try:
                    meta = struct_blit_meta_standalone.parse(meta_head + serial.read(meta_size))
                except ConstructError:
                    pass

            block_size = 64 * 1024

            offset_blocks = offset // block_size
            size_blocks = (size - 1) // block_size + 1

            print(f"""Game at block {offset_blocks}
    Size:        {size_blocks} blocks ({size / 1024:.1f}kB)""")

            if meta is not None:
                print(f"""    Title:       {meta.data.title}
    Description: {meta.data.description}
    Version:     {meta.data.version}
    Author:      {meta.data.author}
""")

            offset_str = serial.read(4)

    @serial_command
    def run_debug(self, serial, args):
        pass

    @serial_command
    def run_reset(self, serial, args):
        logging.info('Resetting your 32Blit...')
        self._reset(serial)

    @serial_command
    def run_info(self, serial, args):
        logging.info('Getting 32Blit run status...')
        status = self._get_status(serial)
        print(f'Running: {status}')



import pathlib
import struct

from elftools.elf.elffile import ELFFile
from elftools.elf.enums import ENUM_RELOC_TYPE_ARM

from ..core.tool import Tool


class Relocs(Tool):
    command = 'relocs'
    help = 'Prepend relocations to a game binary'

    def __init__(self, parser):
        Tool.__init__(self, parser)

        self.parser.add_argument('--bin-file', type=pathlib.Path, help='Input .bin', required=True)
        self.parser.add_argument('--elf-file', type=pathlib.Path, help='Input .elf', required=True)
        self.parser.add_argument('--output', type=pathlib.Path, help='Output file', required=True)

    def run(self, args):
        with open(args.elf_file, 'rb') as f:
            elffile = ELFFile(f)

            # get addresses of GOT values that need patched
            got_offsets = self.get_flash_addr_offsets(elffile.get_section_by_name('.got'))

            # and the init/fini arrays
            init_offsets = self.get_flash_addr_offsets(elffile.get_section_by_name('.init_array'))
            fini_offsets = self.get_flash_addr_offsets(elffile.get_section_by_name('.fini_array'))

            reloc_offsets = []

            # find sidata/sdata
            sidata = 0
            sdata = 0
            relocs = elffile.get_section_by_name('.rel.text')
            symtable = elffile.get_section(relocs['sh_link'])
            for reloc in relocs.iter_relocations():
                symbol = symtable.get_symbol(reloc['r_info_sym'])
                if symbol.name == '_sidata':
                    sidata = symbol['st_value']
                elif symbol.name == '_sdata':
                    sdata = symbol['st_value']

                if sidata and sdata:
                    break

            assert(sidata != 0 and sdata != 0)

            # get all .data relocations
            relocs = elffile.get_section_by_name('.rel.data')
            symtable = elffile.get_section(relocs['sh_link'])

            for reloc in relocs.iter_relocations():
                symbol = symtable.get_symbol(reloc['r_info_sym'])

                if reloc['r_info_type'] != ENUM_RELOC_TYPE_ARM['R_ARM_ABS32']:
                    continue

                # doesn't point to flash
                if symbol['st_value'] < 0x90000000:
                    continue

                # map RAM address back to flash
                flash_offset = (reloc['r_offset'] - sdata) + sidata

                assert((flash_offset & 3) == 0)

                reloc_offsets.append(flash_offset)

            with open(args.output, 'wb') as out_f:
                all_offsets = got_offsets + init_offsets + fini_offsets + reloc_offsets
                all_offsets.sort()

                out_f.write(b"RELO")
                out_f.write(struct.pack("<L", len(all_offsets)))
                for off in all_offsets:
                    out_f.write(struct.pack("<L", off))

                out_f.write(open(args.bin_file, 'rb').read())

    def get_flash_addr_offsets(self, section):
        section_data = section.data()
        section_addrs = struct.unpack(f'<{len(section_data) // 4}I', section_data)
        section_offsets = []

        for i, addr in enumerate(section_addrs):
            # filter out non-flash
            if addr < 0x90000000:
                continue

            # offset to this address in the section
            section_offsets.append(section['sh_addr'] + i * 4)

        return section_offsets



import io
import logging

from bitstring import BitArray
from PIL import Image

from ..core.assetbuilder import AssetBuilder
from ..core.palette import Colour, Palette, type_palette
from ..core.struct import struct_blit_image


class ImageAsset(AssetBuilder):
    command = 'image'
    help = 'Convert images/sprites for 32Blit'
    types = ['image']
    typemap = {
        'image': ('.png', '.gif')
    }

    def __init__(self, parser=None):
        self.options.update({
            'palette': (Palette, Palette()),
            'transparent': Colour,
            'packed': (str, 'yes'),
            'strict': (bool, False)
        })

        AssetBuilder.__init__(self, parser)

        self.palette = None
        self.transparent = None
        self.packed = True
        self.strict = False

        if self.parser is not None:
            self.parser.add_argument('--palette', type=type_palette, default=None, help='Image or palette file of colours to use')
            self.parser.add_argument('--transparent', type=Colour, help='Transparent colour')
            self.parser.add_argument('--packed', type=str, nargs='?', default='yes', choices=('yes', 'no'), help='Pack into bits depending on palette colour count')
            self.parser.add_argument('--strict', action='store_true', help='Reject colours not in the palette')

    def prepare(self, args):
        AssetBuilder.prepare(self, args)

        if type(self.packed) is not bool:
            self.packed = self.packed == 'yes'

        if self.transparent is not None:
            r, g, b = self.transparent
            p = self.palette.set_transparent_colour(r, g, b)
            if p is not None:
                logging.info(f'Found transparent colour ({r},{g},{b}) in palette')
            else:
                logging.warning(f'Could not find transparent colour ({r},{g},{b}) in palette')

    def quantize_image(self, input_data):
        if self.strict and len(self.palette) == 0:
            raise TypeError("Attempting to enforce strict colours with an empty palette, did you really want to do this?")
        # Since we already have bytes, we need to pass PIL an io.BytesIO object
        image = Image.open(io.BytesIO(input_data)).convert('RGBA')
        w, h = image.size
        output_image = Image.new('P', (w, h))
        for y in range(h):
            for x in range(w):
                r, g, b, a = image.getpixel((x, y))
                if self.transparent is not None and (r, g, b) == tuple(self.transparent):
                    a = 0x00
                index = self.palette.get_entry(r, g, b, a, strict=self.strict)
                output_image.putpixel((x, y), index)

        return output_image

    def to_binary(self, input_data):
        image = self.quantize_image(input_data)

        if self.packed:
            bit_length = self.palette.bit_length()
            image_data = BitArray().join(BitArray(uint=x, length=bit_length) for x in image.tobytes()).tobytes()
        else:
            image_data = image.tobytes()

        return struct_blit_image.build({
            'type': 'PK' if self.packed else 'RW',
            'width': image.size[0],
            'height': image.size[1],
            'palette_entries': len(self.palette),
            'palette': self.palette.tostruct(),
            'data': image_data
        })



import argparse
import binascii
import io
import logging
import pathlib
import struct
from datetime import datetime

import yaml
from bitstring import BitArray, BitStream
from construct.core import StreamError
from PIL import Image

from ..asset.image import ImageAsset
from ..core.struct import (blit_game, blit_game_with_meta,
                           blit_game_with_meta_and_relo, blit_icns,
                           struct_blit_image)
from ..core.tool import Tool


class Metadata(Tool):
    command = 'metadata'
    help = 'Tag a 32Blit .bin file with metadata'

    def __init__(self, parser):
        Tool.__init__(self, parser)
        self.parser.add_argument('--config', type=pathlib.Path, help='Metadata config file')
        self.parser.add_argument('--icns', type=pathlib.Path, help='macOS icon output file')
        self.parser.add_argument('--file', type=pathlib.Path, help='Input file')
        self.parser.add_argument('--force', action='store_true', help='Force file overwrite')
        self.parser.add_argument('--dump-images', action='store_true', help='Dump images from metadata')

        self.config = {}

    def parse_config(self, config_file):
        config = open(config_file).read()
        config = yaml.safe_load(config)

        required = ['title', 'description', 'version', 'author']

        for option in required:
            if option not in config:
                raise ValueError(f'Missing required option "{option}" from {config_file}')

        self.config = config

    def prepare_image_asset(self, name, config, working_path):
        image_file = pathlib.Path(config.get('file', ''))
        config['input_file'] = image_file
        config['output_file'] = image_file.with_suffix('.bin')
        if not image_file.is_file():
            image_file = working_path / image_file
        if not image_file.is_file():
            raise ValueError(f'{name} "{image_file}" does not exist!')
        asset = ImageAsset(argparse.ArgumentParser().add_subparsers())
        asset.prepare(config)

        return asset.to_binary(open(image_file, 'rb').read())

    def packed_to_image(self, image):
        num_pixels = image.width * image.height
        image_icon_data = BitArray().join(BitArray(uint=x, length=8) for x in image.data)
        image_icon_data = BitStream(image_icon_data).readlist(",".join(f"uint:{image.bit_length}" for _ in range(num_pixels)))

        raw_data = bytes()
        for i in image_icon_data:
            rgba = image.palette[i]
            raw_data += bytes([
                rgba.r,
                rgba.g,
                rgba.b,
                rgba.a,
            ])

        return Image.frombytes("RGBA", (image.width, image.height), raw_data)

    def binary_size(self, bin):
        return struct.unpack('<I', bin[16:20])[0] & 0xffffff

    def checksum(self, bin):
        return struct.pack('<I', binascii.crc32(bin))

    def build_icns(self, config):
        image_file = pathlib.Path(config.get('file', ''))
        if not image_file.is_file():
            raise ValueError(f'splash "{image_file}" does not exist!')

        blit_icon = Image.frombytes("P", (22, 22), b'\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x00\x00\x01\x01\x01\x01\x01\x01\x01\x01\x02\x02\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x00\x00\x01\x01\x01\x01\x01\x01\x01\x01\x02\x02\x00\x00\x00\x00\x00\x00\x00\x00\x01\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x02\x02\x00\x00\x00\x00\x00\x00\x00\x00\x01\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x02\x02\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x00\x00\x01\x01\x01\x01\x01\x01\x01\x01\x02\x02\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x00\x00\x01\x01\x01\x01\x01\x01\x01\x01\x02\x02\x00\x00\x00\x00\x00\x00\x00\x00\x01\x01\x00\x00\x01\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01\x01\x00\x00\x01\x01\x00\x00\x00\x00\x00\x00\x00\x00\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x00\x00\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x00\x00\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01\x01\x00\x00\x00\x00\x00\x00\x00\x00\x01\x01\x00\x00\x02\x02\x00\x00\x01\x01\x00\x00\x01\x01\x00\x00\x00\x00\x00\x00\x00\x00\x01\x01\x00\x00\x02\x02\x00\x00\x01\x01\x00\x00\x01\x01\x01\x01\x01\x01\x01\x01\x00\x00\x01\x01\x00\x00\x00\x00\x00\x00\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x00\x00\x01\x01\x00\x00\x00\x00\x00\x00\x01\x01\x01\x01\x02\x02\x00\x00\x00\x00\x01\x01\x00\x00\x01\x01\x00\x00\x01\x01\x00\x00\x01\x01\x00\x00\x02\x02\x00\x00\x00\x00\x01\x01\x00\x00\x01\x01\x00\x00\x01\x01\x00\x00\x01\x01\x00\x00\x02\x02\x00\x00\x00\x00\x01\x01\x00\x00\x01\x01\x00\x00\x01\x01\x00\x00\x01\x01\x00\x00\x02\x02\x00\x00\x00\x00\x01\x01\x00\x00\x01\x01\x00\x00\x01\x01\x00\x00\x01\x01\x00\x00\x02\x02\x01\x01\x01\x01\x01\x01\x00\x00\x01\x01\x00\x00\x01\x01\x00\x00\x01\x01\x01\x01\x02\x02\x01\x01\x01\x01\x01\x01\x00\x00\x01\x01\x00\x00\x01\x01\x00\x00\x01\x01\x01\x01')
        blit_icon.putpalette([0, 0, 0, 255, 255, 255, 0, 255, 0])
        blit_icon = blit_icon.convert("RGBA")

        source = Image.open(image_file).convert("RGBA")

        image = Image.new("RGB", (128, 128), (0, 0, 0))

        image.paste(source, (0, 0))
        image.paste(blit_icon, (101, 101), blit_icon)

        image_bytes = io.BytesIO()
        image.convert("RGB").save(image_bytes, format="PNG")
        image_bytes.seek(0)

        del image

        return blit_icns.build({'data': image_bytes.read()})

    def run(self, args):
        if args.file is None and args.config is None:
            self.parser.error('the following arguments are required: --config and/or --file')

        if args.file and not args.file.is_file():
            raise ValueError(f'Unable to find bin file at {args.file}')

        icon = b''
        splash = b''
        working_path = pathlib.Path('')

        game = None

        if args.file:
            bin = open(args.file, 'rb').read()
            try:
                game = blit_game.parse(bin)
            except StreamError:
                raise ValueError(f'Invalid 32blit binary file {args.file}')

        # No config supplied, so dump the game info
        if args.config is None:
            print(f"""
Parsed:      {args.file.name} ({game.bin.length:,} bytes)""")
            if game.relo is not None:
                print(f"""Relocations: Yes ({len(game.relo.relocs)})""")
            else:
                print("""Relocations: No""")
            if game.meta is not None:
                print(f"""Metadata:    Yes

    Title:       {game.meta.data.title}
    Description: {game.meta.data.description}
    Version:     {game.meta.data.version}
    Author:      {game.meta.data.author}
""")
                if game.meta.data.icon is not None:
                    game_icon = game.meta.data.icon
                    print(f"""    Icon:        {game_icon.width}x{game_icon.height} ({len(game_icon.palette)} colours)""")
                    if args.dump_images:
                        image_icon = self.packed_to_image(game_icon)
                        image_icon_filename = args.file.with_suffix(".icon.png")
                        image_icon.save(image_icon_filename)
                        print(f"    Dumped to:   {image_icon_filename}")
                if game.meta.data.splash is not None:
                    game_splash = game.meta.data.splash
                    print(f"""    Splash:      {game_splash.width}x{game_splash.height} ({len(game_splash.palette)} colours)""")
                    if args.dump_images:
                        image_splash = self.packed_to_image(game_splash)
                        image_splash_filename = args.file.with_suffix(".splash.png")
                        image_splash.save(image_splash_filename)
                        print(f"    Dumped to:   {image_splash_filename}")
                print("")
            else:
                print("""Metadata:    No
""")
            return

        if args.config.is_file():
            working_path = args.config.parent
            self.parse_config(args.config)
            logging.info(f'Using config at {args.config}')
        else:
            logging.warning(f'Unable to find metadata config at {args.config}')

        if 'icon' in self.config:
            icon = self.prepare_image_asset('icon', self.config['icon'], working_path)
        else:
            raise ValueError('An 8x8 pixel icon is required!"')

        if 'splash' in self.config:
            splash = self.prepare_image_asset('splash', self.config['splash'], working_path)
            if args.icns is not None:
                if not args.icns.is_file() or args.force:
                    open(args.icns, 'wb').write(self.build_icns(self.config['splash']))
                    logging.info(f'Saved macOS icon to {args.icns}')
        else:
            raise ValueError('A 128x96 pixel splash is required!"')

        if not game:
            return

        title = self.config.get('title')
        description = self.config.get('description')
        version = self.config.get('version')
        author = self.config.get('author')

        if len(title) > 24:
            raise ValueError('Title should be a maximum of 24 characters!"')

        if len(description) > 128:
            raise ValueError('Description should be a maximum of 128 characters!')

        if len(version) > 16:
            raise ValueError('Version should be a maximum of 16 characters! eg: "v1.0.2"')

        if len(author) > 16:
            raise ValueError('Author should be a maximum of 16 characters!')

        if game.meta is not None:
            if not args.force:
                logging.critical(f'Refusing to overwrite metadata in {args.file}')
                return 1

        game.meta = {
            'data': {
                'date': datetime.now().strftime("%Y%m%dT%H%M%S"),
                'title': title,
                'description': description,
                'version': version,
                'author': author,
                'icon': struct_blit_image.parse(icon),
                'splash': struct_blit_image.parse(splash)
            }
        }

        # Force through a non-optional builder if relo symbols exist
        # since we might have modified them and want to hit parser errors
        # if something has messed up
        if game.relo is not None:
            bin = blit_game_with_meta_and_relo.build(game)
        else:
            bin = blit_game_with_meta.build(game)

        logging.info(f'Adding metadata to {args.file}')
        open(args.file, 'wb').write(bin)

        return 0



import logging
import pathlib
import re

from .outputformat import (CHeader, OutputFormat, output_formats,
                           parse_output_format)
from .tool import Tool


class AssetBuilder(Tool):
    formats = output_formats
    no_output_file_default_format = CHeader

    options = {
        'input_file': pathlib.Path,
        'input_type': str,
        'output_file': pathlib.Path,
        'output_format': parse_output_format,
        'symbol_name': str,
        'force': bool,
        'prefix': str,
        'working_path': pathlib.Path
    }

    def __init__(self, parser=None):
        Tool.__init__(self, parser)

        if self.parser is not None:
            self.parser.add_argument('--input_file', type=pathlib.Path, required=True, help='Input file')
            if(len(self.types) > 1):
                self.parser.add_argument('--input_type', type=str, default=None, choices=self.types, help='Input file type')
            self.parser.add_argument('--output_file', type=pathlib.Path, default=None)
            self.parser.add_argument('--output_format', type=str, default=None, choices=self.formats.keys(), help='Output file format')
            self.parser.add_argument('--symbol_name', type=str, default=None, help='Output symbol name')
            self.parser.add_argument('--force', action='store_true', help='Force file overwrite')

    def prepare(self, opts):
        """Imports a dictionary of options to class variables.

        Requires options to already be in their correct types.

        """
        Tool.prepare(self, opts)

        if self.symbol_name is None:
            if self.working_path is None:
                name = '_'.join(self.input_file.parts)
            else:
                name = '_'.join(self.input_file.relative_to(self.working_path).parts)
            name = name.replace('.', '_')
            name = re.sub('[^0-9A-Za-z_]', '_', name)
            self.symbol_name = name.lower()

        if type(self.prefix) is str:
            self.symbol_name = self.prefix + self.symbol_name

        if self.output_format is None:
            self._guess_format()
        elif type(self.output_format) is str:
            self.output_format = parse_output_format(self.output_format)
        elif not issubclass(self.output_format, OutputFormat):
            raise ValueError(f'Invalid format {self.output_format}, choices {self.formats.keys()}')

        if self.input_type is None:
            self._guess_type()
        elif self.input_type not in self.types:
            raise ValueError(f'Invalid type {self.input_type}, choices {self.types}')

    def run(self, args):
        self.prepare_options(vars(args))

        output_data = self.build()

        self.output(output_data, self.output_file, self.output_format, self.force)

    def prepare_options(self, opts):
        """Imports a dictionary of options to class variables.

        Used for external callers which don't invoke `run` on this class.

        Converts all options into their correct types via the specified validators.

        """
        if self.options is not None and opts is not None:
            for option_name, option_value in opts.items():
                if option_name in self.options:
                    option_type = self.options[option_name]
                    default_value = None
                    if type(option_type) is tuple:
                        option_type, default_value = option_type
                    if option_value is not None:
                        opts[option_name] = option_type(option_value)
                    else:
                        opts[option_name] = default_value
                else:
                    logging.info(f'Ignoring unsupported {self.command} option {option_name}')

        self.prepare(opts)

    def build(self):
        input_data = open(self.input_file, 'rb').read()

        output_data = self.to_binary(input_data)

        return self.output_format().build(output_data, self.symbol_name)

    def _guess_type(self):
        for input_type, extensions in self.typemap.items():
            for extension in extensions:
                if self.input_file.name.endswith(extension):
                    self.input_type = input_type
                    logging.info(f"Guessed type {input_type} for {self.input_file}")
                    return

        raise TypeError(f"Unable to identify type of input file {self.input_file}")

    def _guess_format(self):
        if self.output_file is None:
            self.output_format = self.no_output_file_default_format
            logging.warning(f"No --output given, writing to stdout assuming {self.no_output_file_default_format.name}")
            return

        for format_name, format_class in self.formats.items():
            for extension in format_class.extensions:
                if self.output_file.name.endswith(extension):
                    self.output_format = format_class
                    logging.info(f"Guessed output format {format_class.name} for {self.output_file}")
                    return

        raise TypeError(f"Unable to identify type of output file {self.output_file}")



class OutputFormat():
    name = 'none'
    components = None
    extensions = ()

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.name

    def build(self, input_data, symbol_name):
        if self.components is None:
            return self.output(input_data, symbol_name)
        else:
            output_components = {}
            for extension in self.components:
                output_components[extension] = getattr(self, f'output_{extension}')(input_data, symbol_name)
            return output_components


class CHeader(OutputFormat):
    name = 'c_header'
    extensions = ('.hpp', '.h')

    def __init__(self):
        OutputFormat.__init__(self)

    def _helper_raw_to_c_source_hex(self, input_data):
        if type(input_data) is str:
            input_data = bytes(input_data, encoding='utf-8')
        return ', '.join([f'0x{c:02x}' for c in input_data])

    def output(self, input_data, symbol_name):
        input_data = self._helper_raw_to_c_source_hex(input_data)
        return f'''inline const uint8_t {symbol_name}[] = {{{input_data}}};
inline const uint32_t {symbol_name}_length = sizeof({symbol_name});'''

    def join(self, ext, filename, data):
        if type(data) is list:
            data = '\n'.join(data)
        return f'''// Auto Generated File - DO NOT EDIT!
#pragma once
#include <cstdint>
{data}
'''


class CSource(CHeader):
    name = 'c_source'
    components = ('hpp', 'cpp')
    extensions = ('.cpp', '.c')

    def __init__(self):
        OutputFormat.__init__(self)

    def output_hpp(self, input_data, symbol_name):
        return f'''extern const uint8_t {symbol_name}[];
extern const uint32_t {symbol_name}_length;'''

    def output_cpp(self, input_data, symbol_name):
        input_data = self._helper_raw_to_c_source_hex(input_data)
        return f'''const uint8_t {symbol_name}[] = {{{input_data}}};
const uint32_t {symbol_name}_length = sizeof({symbol_name});'''

    def join(self, ext, filename, data):
        if type(data) is list:
            data = '\n'.join(data)
        if ext == 'cpp':
            header = filename.with_suffix('.hpp').name
            return f'''// Auto Generated File - DO NOT EDIT!
#include "{header}"
{data}
'''
        else:
            return f'''// Auto Generated File - DO NOT EDIT!
#pragma once
#include <cstdint>
{data}
'''


class RawBinary(OutputFormat):
    name = 'raw_binary'
    extensions = ('.raw', '.bin')

    def __init__(self):
        OutputFormat.__init__(self)

    def output(self, input_data, symbol_name):
        return input_data

    def join(self, ext, filename, data):
        if type(data) is list:
            data = b''.join(data)
        return data


output_formats = {}
name = None
object = None


def parse_output_format(value):
    if type(value) is str:
        return output_formats.get(value, None)
    if issubclass(value, OutputFormat):
        return value
    return output_formats[value]


for name, object in globals().items():
    try:
        if issubclass(object, OutputFormat) and not name == 'OutputFormat':
            output_formats[object.name] = object
    except TypeError:
        pass
